import {ref, Ref} from "vue"
import {LocationQuery, LocationQueryValue} from "vue-router"

export class ScrollListKey {

  // 리스트를 강제로 갱신하는 경우 값을 하나씩 올려 사용한다.
  private forceUpdate = ref(0)

  constructor(private params?: Ref[]) {
  }

  setParams(params: Ref[]) {
    this.params = params
  }

  forceRefresh(): void {
    this.forceUpdate.value++
  }

  key(): string {
    const key = this.params
        ?.map(param => param.value)
        .map(value => value ? value.toString() : '')
        .join('_') || ''
    return key + (this.forceUpdate.value > 0 ? `_${this.forceUpdate.value}` : '')
  }
}

export type QueryConverter<T> = {
  toQueryString?: (value: T) => string
  fromQueryString?: (query: LocationQueryValue | LocationQueryValue[]) => T | undefined
}

export type ScrollListFilterItem = {
  name: string
  value: Ref
  defaultValue?: unknown
  queryConverter?: QueryConverter<unknown>
}

export class ScrollListFilter {

  private queryRef = ref('')

  constructor(private params?: ScrollListFilterItem[]) {
  }

  setParams(params: ScrollListFilterItem[]) {
    this.params = params
  }

  createUrlQuery(locationQuery: LocationQuery): LocationQuery {
    const nonFilter = this.extractNonFilterQuery(locationQuery)
    const filter = this.createRequestQuery()
    return {...nonFilter, ...filter}
  }

  static isQueryEqual(query1: LocationQuery, query2: LocationQuery): boolean {
    return JSON.stringify(query1, Object.keys(query1).sort()) === JSON.stringify(query2, Object.keys(query2).sort())
  }

  private createRequestQuery(): LocationQuery {
    const query: LocationQuery = {}
    this.params
        ?.filter(param => param.value.value && (param.value.value !== param.defaultValue))
        .forEach(param => {
          const queryValue = this.convertToQuery(param)
          if (queryValue) query[param.name] = queryValue
        })
    return query
  }

  private extractNonFilterQuery(query: LocationQuery): LocationQuery {
    const filterNames = this.params?.map(param => param.name) || []
    return Object.keys(query || [])
        .reduce<LocationQuery>((acc, current) => {
          if (!filterNames.includes(current))
            acc[current] = query?.[current] || ''
          return acc
        }, {})
  }

  syncUrlQuery(locationQuery: LocationQuery): void {
    this.params?.forEach(param => {
      const queryValue = locationQuery[param.name]
      param.value.value = this.convertFromQuery(param, queryValue)
    })
    const query = this.createRequestQuery()
    this.queryRef.value = Object.keys(query).length == 0 ? '' : JSON.stringify(query)
  }

  private convertToQuery(param: ScrollListFilterItem): string | undefined {
    const value = param.value.value
    if (!value || (JSON.stringify(value) === JSON.stringify(param.defaultValue))) {
      return undefined // value 가 없거나 defaultValue 와 같다면 undefined
    } else if (param.queryConverter?.toQueryString) {
      return param.queryConverter.toQueryString(value) // queryConverter 가 존재하면 toQueryString 결과
    } else {
      return value.toString() // 그렇지 않으면 toString()
    }
  }

  private convertFromQuery(param: ScrollListFilterItem,
                           queryValue?: LocationQueryValue | LocationQueryValue[]): unknown | undefined {
    if (!queryValue) {
      return param.defaultValue // queryValue 가 없으면 defaultValue
    } else if (param.queryConverter?.fromQueryString) {
      return param.queryConverter?.fromQueryString(queryValue) // queryConverter 가 존재하면 fromQueryString 결과
    } else {
      return String(queryValue) // 그렇지 않으면 string
    }
  }

  createListKey(): ScrollListKey {
    return new ScrollListKey([this.queryRef])
  }

  getParamValues(): Record<string, Ref> {
    const result: Record<string, Ref> = {}
    this.params?.forEach(param => result[param.name] = (param.value || param.defaultValue))
    return result
  }
}

export const numberQueryConverter: QueryConverter<unknown> = {

  toQueryString(value: unknown): string {
    return String(value)
  },

  fromQueryString(query: LocationQueryValue | LocationQueryValue[]): unknown | undefined {
    const value = Array.isArray(query) ? String(query[0]) : String(query)
    return Number(value)
  }
}