import "vue-loading-overlay/dist/vue-loading.css"
import {ActiveLoader, Props, useLoading} from "vue-loading-overlay"

const defaultProps: Props = {
  color: '#3f51b5',
  loader: 'spinner' // spinner, dots, bars
}

const inBrowser = (typeof process === 'undefined')

// show()/hide() 호출이 중첩되는 경우를 처리하기 위해, 호출 카운트를 관리한다.
class LoadingOverlayGuard {
  loadingOverlay = useLoading(defaultProps)
  showCount = 0
  loader?: ActiveLoader

  show(): void {
    // vue test-utils 에서 $refs 가 비어있는 경우가 있어 hide() 시 오류가 발생하는 경우가 있음
    // 이를 막기 위해 node 에서는 loading-overlay 를 표시하지 않는다.
    if (inBrowser) {
      if (this.showCount <= 0) {
        this.loader = this.loadingOverlay.show()
      }
      ++this.showCount
    }
  }

  hide() {
    if (this.showCount > 0) {
      --this.showCount
      if ((this.showCount <= 0) && this.loader) {
        this.loader.hide()
        this.loader = undefined
      }
    }
  }
}

const defaultLoadingOverlayGuard = new LoadingOverlayGuard()

/**
 * `func` 함수 호출 전에 LoadingOverlay 를 show() 하고,
 * 호출이 끝난 후(Promise 가 모두 resolve 된 후)에 hide() 한다.
 *
 * 만약 중첩되어 show() 된 경우, 마지막 hide() 시점에 실제로 hide 된다.
 *
 * `func` 함수는 async 함수이어야 한다. async 하지 않은 함수 호출에 LoadingOverlay 를 표시할 수는 없다.
 *
 * @param func 호출할 함수.
 */
export const withLoading = async <T>(func: () => T): Promise<T> => {
  try {
    defaultLoadingOverlayGuard.show()
    return await func()
  } finally {
    defaultLoadingOverlayGuard.hide()
  }
}

/**
 * LoadingOverlay 를 표시한다.
 *
 * 대부분은 `withLoading()` 을 사용할 것이고, 이 함수를 사용하는 경우는 극히 드물 것이다.
 */
export const showLoading = (): void => {
  defaultLoadingOverlayGuard.show()
}

/**
 * LoadingOverlay 를 감춘다.
 *
 * 대부분은 `withLoading()` 을 사용할 것이고, 이 함수를 사용하는 경우는 극히 드물 것이다.
 */
export const hideLoading = (): void => {
  defaultLoadingOverlayGuard.hide()
}