/**
 * Lazy
 * ```ts
 * import lazy from "@/lib/directive/lazy.ts"
 * new Vue({
 *   directives: {lazy}
 * })
 * ```
 * * enabled (default: true)
 * * onIntersect (default: null)
 * * once (default: true)
 *
 * ```html
 * <UiImage v-lazy="{enabled:true, onIntersect: el => el.src = computedSrc, once: true}"/>
 * <UiImage v-lazy="{enabled:true, onIntersect: (el, vnode) => ..., once: false}"/>
 * ```
 */

import {Directive, DirectiveBinding, VNode} from "vue"
import {supportIntersectionObserver} from "@/config/clientContext"

export type OnIntersect = (el: Element, vnode: VNode) => void

declare global {
  interface Element {
    onIntersect?: () => void
    onIntersectOnce?: boolean
  }
}

/**
 * 디폴트 IntersectionObserver. view port 를 root 로 한다.
 */
const defaultIntersectionObserver = !supportIntersectionObserver
    ? undefined
    : new IntersectionObserver((entries, observer) => {
      entries.forEach(entry => {
        if (entry.isIntersecting) { // 이미지가 교차영역에 진입 할 때
          entry.target.onIntersect?.()
          if (entry.target.onIntersectOnce) {
            observer.unobserve(entry.target) // 로딩이 끝난 후 관찰을 하지 않는다.
            delete entry.target.onIntersect
          }
        }
      })
    })


export default {
  mounted: (element: Element, {value}: DirectiveBinding, vnode: VNode): void => {

    const enabled = (typeof value === 'object') ? !!value.enabled : !!value
    const onIntersectHandler = (typeof value === 'object') ? value.onIntersect as OnIntersect : null
    const once = (typeof value === 'object' && value.once !== undefined) ? value.once : true

    // 브라우저에서 IntersectionObserver 를 지원하는 경우에만 동작하도록 한다.
    // 브라우저에서 지원하지 않으면, handler 를 곧바로 실행한다.
    if (defaultIntersectionObserver && (enabled === undefined || enabled)) {
      element.onIntersect = () => onIntersectHandler?.(element, vnode)
      element.onIntersectOnce = once
      defaultIntersectionObserver.observe(element)
    } else {
      onIntersectHandler?.(element, vnode)
    }
  }
} as Directive